#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*Listing of local function templates.*/
void    end_of_page (FILE *fp);       /*signal end-of-page*/
void    print_line (FILE *fp);        /*print line in buffer to fp*/
void    ps_header (FILE *fp);         /*print PS header to fp*/
int     read_line (FILE *fp);         /*read a line from fp into buffer*/
void    reset_page (FILE *fp);        /*reset cursor to upper-left of page*/
void    send_eot (FILE *fp);          /*signal end-of-file*/
void    eojob (FILE *fp);             /*signal end-of-job & cleanup things*/
void    show_usage (void);            /*show usage of filter*/


/***********Listing of C library functions used*****************
    exit ()       ==> exit to OS.
    fclose ()     ==> close an open file stream.
    feof ()       ==> check for the end-of-file condition.
    fgetc ()      ==> get a character from a file stream.
    fopen ()      ==> open a file stream.
    fprintf ()    ==> print formatted output to a file stream.
    strlen ()     ==> return the length of a character string.
****************************************************************/

/*Global variables.*/
char    buffer[BUFSIZ];     /*Global buffer to hold input.*/
FILE    *input;             /*Pointers to the input     */
FILE    *output;            /*  and output file streams.*/
int     nlines;             /*Global line counter.*/

/*Global definitions.*/
#define BACKSLASH           (char)92
#define LEFTPARENTHESIS     (char)40
#define RIGHTPARENTHESIS    (char)41
#define TAB                 (char)9
#define SPACE               (char)32
#define FORMFEED            (char)12

#define MAXLINESPERPAGE     56
#define MAXCHARSPERLINE     93
#define TABWIDTH            8


void    show_usage ()
{
    fprintf (stderr, "*********************************************\n");
    fprintf (stderr, "*                                           *\n");
    fprintf (stderr, "* asc2ps:   ASCII to PostScript Translator. *\n");
    fprintf (stderr, "*                                           *\n");
    fprintf (stderr, "* Syntax:                                   *\n");
    fprintf (stderr, "*                                           *\n");
    fprintf (stderr, "*    asc2ps   input-file   [output-file]    *\n");
    fprintf (stderr, "*                                           *\n");
    fprintf (stderr, "* If \"output-file\" is not specified, the    *\n");
    fprintf (stderr, "* output is routed to the standard output   *\n");
    fprintf (stderr, "* device, i.e., the console.                *\n");
    fprintf (stderr, "*                                           *\n");
    fprintf (stderr, "*********************************************\n");
    exit (1);
}

/*The ps_header function must be called once at the start of a print job.*/

void    ps_header (FILE *f)
{
    fprintf (f, "%%!ASCII to PS filter\n");
    fprintf (f, "%%Courtesy of C. Perez Ramos\n");
    fprintf (f, "%% ---- TEXT HANDLING FUNCTIONS ----\n");
    fprintf (f, "/inch { 72 mul } def\n");
    fprintf (f, "/AdvanceLine { /NextLine NextLine LineSpace sub def } def\n");
    fprintf (f, "/showline { LeftMargin NextLine moveto show } def\n");
    fprintf (f, "%% ---- PAGE LAYOUT CONSTANTS ----\n");
    fprintf (f, "/LeftStart  .5 inch def\n");
    fprintf (f, "/TopStart   10 inch def\n");
    fprintf (f, "/LineSpace  12 def\n");
    fprintf (f, "%% ---- FONT DEFINITION ----\n");
    fprintf (f, "/Courier findfont\n10 scalefont\nsetfont\n");
    fprintf (f, "\045\045 ---- END OF PS HEADER ----\n\n");
    fprintf (f, "gsave\n");
}

/*Reset the pointer to the top left corner of the page.*/

void    reset_page (FILE *f)
{
    extern  int nlines;

    fprintf (f, "\045\045  ---- MOVE POINTER TO THE TOP OF PAGE ----\n");
    fprintf (f, "/NextLine TopStart def\n");
    fprintf (f, "/LeftMargin LeftStart def\n");
    fprintf (f, "\045\045  ---- INSERT LINES TO BE PRINTED HERE ----\n");
    nlines = 0;
}

/*Send the PostScript EOT marker (ASCII 4) to the file stream f.*/

void    send_eot (FILE *f)
{
    fprintf (f, "%c\n", 4);
}

/*The eoj function should be called at the end of every job*/
/*to cleanup things.                                       */

void    eojob (FILE *f)
{
    fprintf (f, "grestore\n");
}

/*Send a PostScript showpage operator at the end of a page.*/

void    end_of_page (FILE *f)
{
    fprintf (f, "showpage\n");
}

/*Read a line of input from the file stream f. The function retrieves*/
/*characters from the file stream until a newline character is found,*/
/*or the end-of-file is found, or the maximum buffer size is reached.*/
/*Special characters (e.g., back slash, left and right parenthesis,  */
/*and tab) are given special attention since they are significant to */
/*the PostScript language.*/

int     read_line (FILE *f)
{
    extern  char buffer[];
    int     c, length, i, n;

    length = 0;

    n = 0;

    c = fgetc (f);

    /*Read characters from the input file until the maximum length of*/
    /*the buffer is exceeded (BUFSIZ), or the end of the file is     */
    /*reached, or a newline character (i.e., carriage return) is read*/

    while ((length < BUFSIZ - 1) && !feof(f) && c != '\n')
    {
        switch (c)
        {
        /*Backslashes and parentheses need special handling within.*/
        /*PostScript strings.*/

            case BACKSLASH:

                buffer[n++] = BACKSLASH;
                buffer[n++] = BACKSLASH;
                ++length;
                break;

            case LEFTPARENTHESIS:

                buffer[n++] = BACKSLASH;
                buffer[n++] = LEFTPARENTHESIS;
                ++length;
                break;

            case RIGHTPARENTHESIS:

                buffer[n++] = BACKSLASH;
                buffer[n++] = RIGHTPARENTHESIS;
                ++length;
                break;

            case TAB:

                if (length < (BUFSIZ - TABWIDTH))  /*Expand tabs to spaces.*/
                {
                    for (i = 0; i < TABWIDTH; ++i)
                        buffer[n++] = SPACE;
                }
                length += TABWIDTH;
                break;

            case FORMFEED:

                buffer[n] = '\0';
                return (FORMFEED);

            default:

                buffer[n++] = (char) c;
                ++length;
        }

        c = fgetc (f);
    }

    buffer[n] = '\0';  /*Terminate buffer with null character.*/

    /*Check if the line length is greater than MAXCHARSPERLINE. If so,*/
    /*print a warning message to stderr, and print the offending line.*/

    if (length > MAXCHARSPERLINE)
    {
        fprintf (stderr, "Warning: long line. May be truncated.\n");
        fprintf (stderr, "==> %s\n", buffer);
    }

    return (0);
}

/*Send a PostScript line to the file stream f. If ASCII character 12*/
/*is the last character in the line, do a form feed.*/

void    print_line (FILE *f)
{
    extern char buffer[];
    extern int  nlines;

    long        lastchar;

    lastchar = strlen(buffer) - 1;

    fprintf (f, "(%s) showline\nAdvanceLine\n", buffer);

    if (buffer[lastchar] == 12)
    {
        nlines = 0;
        reset_page (f);
        end_of_page (f);
    }
}


/*Entry point for ASC2PS program.*/

void    main (int argc, char *argv[])
{
    extern  FILE    *input;
    extern  FILE    *output;
    extern  int      nlines;
            int      result;

    /*Initialize the input file as stdin (keyboard). Initialize the*/
    /*output file as stdout (console screen).*/

    input = stdin;
    output = stdout;

    switch (argc - 1)
    {
        case 1:

            /*One argument given. Assume it is name of input file.*/

            if ((input = fopen (argv[1], "r")) == NULL)
            {
                fprintf (stderr, "File %s not found.\n", argv[1]);
                exit (1);
            }
            break;

        case 2:

            /*Two arguments given. Assume they're input & output files.*/

            if ((input = fopen (argv[1], "r")) == NULL)
            {
                fprintf (stderr, "File %s not found.\n", argv[1]);
                exit (1);
            }

            if ((output = fopen (argv[2], "wt")) == NULL)
            {
                fprintf (stderr, "Error opening %s for output.\n", argv[2]);
                fclose (input);
                exit (1);
            }
            break;

        default:

            /*If neither 1 or 2 arguments were given, print a message*/
            /*showing the usage of the program and exit to OS.*/

            show_usage ();
            break;
    }

    /*Send an eot signal at the start to terminate any jobs that failed*/
    /*to exit gracefully prior to executing the job. This option should*/
    /*be "commented out" on certain machines which require that the    */
    /*first character in a PostScript file be "%!". To "comment out"   */
    /*this option, enclose the following line with "comment" delimiters*/

    send_eot (output);

    /*Print the PostScript header to the output file stream. This step*/
    /*is necessary if any output is to be generated.*/

    ps_header (output);


    /*Loop while the end-of-file has not been reached, and print all*/
    /*lines in the input file stream to the output file stream.*/

    while (!feof (input))
    {
        reset_page (output);

        do
        {
            result = read_line (input);

            switch (result)
            {
                case FORMFEED:

                    if (strlen (buffer) != 0) print_line  (output);
                    end_of_page (output);
                    reset_page (output);
                    break;

                default:

                    print_line (output);
                    ++nlines;
                    break;
            }
        }
        while ((nlines < MAXLINESPERPAGE) && !feof (input));

        end_of_page (output);
    }

    eojob (output);

    /*All done. Send an eot signal to the output file stream.*/

    send_eot (output);

    /*Close the input and output file streams.*/

    fclose (input);

    fclose (output);
}

